/*
 * @(#)GIFEncoder.java    0.90 4/21/96 Adam Doppelt
 */

package JQJoiner.GIFEncoderPkg;

import java.io.*;
import java.awt.*;
import java.awt.image.*;
import JQJoiner.GIFDecoderPkg.*;

/**
 * GIFEncoder is a class which takes an image and saves it to a stream
 * using the GIF file format (<A
 * HREF="http://www.dcs.ed.ac.uk/%7Emxr/gfx/">Graphics Interchange
 * Format</A>). A GIFEncoder
 * is constructed with either an AWT Image (which must be fully
 * loaded) or a set of RGB arrays. The image can be written out with a
 * call to <CODE>Write</CODE>.<P>
 *
 * Three caveats:
 * <UL>
 *   <LI>GIFEncoder will convert the image to indexed color upon
 *   construction. This will take some time, depending on the size of
 *   the image. Also, actually writing the image out (Write) will take
 *   time.<P>
 *
 *   <LI>The image cannot have more than 256 colors, since GIF is an 8
 *   bit format. For a 24 bit to 8 bit quantization algorithm, see
 *   Graphics Gems II III.2 by Xialoin Wu. Or check out his <A
 *   HREF="http://www.csd.uwo.ca/faculty/wu/cq.c">C source</A>.<P>
 *
 *   <LI>Since the image must be completely loaded into memory,
 *   GIFEncoder may have problems with large images. Attempting to
 *   encode an image which will not fit into memory will probably
 *   result in the following exception:<P>
 *   <CODE>java.awt.AWTException: Grabber returned false: 192</CODE><P>
 * </UL><P>
 *
 * GIFEncoder is based upon gifsave.c, which was written and released
 * by:<P>
 * <CENTER>
 *                                  Sverre H. Huseby<BR>
 *                                   Bjoelsengt. 17<BR>
 *                                     N-0468 Oslo<BR>
 *                                       Norway<P>
 *
 *                                 Phone: +47 2 230539<BR>
 *                                 sverrehu@ifi.uio.no<P>
 * </CENTER>
 * @version 0.90 21 Apr 1996
 * @author <A HREF="http://www.cs.brown.edu/people/amd/">Adam Doppelt</A> */
public class GIFEncoder {
    short width_, height_;
    int numColors_;
    byte pixels_[];
    byte colors_[] = null;
    
    OutputStream output_;
    
    ScreenDescriptor sd_;
    ImageDescriptor id_;

/**
 * Construct a GIFEncoder. The constructor will convert the image to
 * an indexed color array. <B>This may take some time.</B><P>
 * 
 * @param image The image to encode. The image <B>must</B> be
 * completely loaded.
 * @exception AWTException Will be thrown if the pixel grab fails. This
 * can happen if Java runs out of memory. It may also indicate that the image
 * contains more than 256 colors.
 *   ----- this version preserves your colorModel ----------
 * */
    
  public GIFEncoder(int width, int height, RGBxSizeSequence colors) throws AWTException   /* NEW */
    // For image concatenation:
    // 1. Call this constructor.  The first two args are for the size of the output.
    //                            The third is obvious.  All components are expected to have the same colormap.
    // 2. Call GIFEncoder.JavaQuatWriteHeader(OutputStream output, String JQArgs, boolean interlaced_p)
    // 3. Call GIFEncoder.JavaQuatWritePixels (byte[] pixels, int first, int bound_exclusive) to write to the
    //                            output file the bytes pixels[first], ..., pixels[bound_exclusive - 1].
    // 4. Call GIFEncoder.JavaQuatWriteFooter().
  { 
    //IndexColorModel colorModel = null; // FOR DEBUG WHEN PARAMETER IS COMMENTED OUT.
  
    width_ = (short)width;
    height_ = (short)height;

            
    // Create the color map for the Encoder, in variable colors_
    int size = colors.byteTable.length;
    //System.out.println("GIFEncoder.init size=" + size);
    size = Math.min(256, size / 3);
    //System.out.println("GIFEncoder.init size=" + size);
    colors_ = new byte[3 * 256];
    
    // Assign values to colorMap colors_: r[0], g[0], b[0], r[1]....
    int i;
    for (i = 0; i < size * 3; i++) {
      //System.out.println("GIFEncoder.init i=" + i);
      colors_[i] = colors.byteTable[i];
    }
    for (; i < 256 * 3; i += 3) {
      //System.out.println("GIFEncoder.init pt2 i=" + i);
      colors_[i] = 0;
      colors_[i + 1] = 0;
      colors_[i + 2] = 0;
    }
//    colors_[0] = (byte)1; //So color 0 is not same as color 1         
     
    int colornum = 256;
    numColors_ = 1 << BitUtils.BitsNeeded(colornum);
  }


/**
 * Writes the image out to a stream in the GIF file format. This will
 * be a single GIF89a image, non-interlaced, with no background color.
 * <B>This may take some time.</B><P>
 *
 * @param output The stream to output to. This should probably be a
 * buffered stream.
 *
 * @exception IOException Will be thrown if a write operation fails.
 * */
    public void javaQuatWriteHeader(OutputStream output, String JQArgs, boolean interlaced_p,
                                    int transparentPixelIndex) /* NEW  -1 for none. */
                                   throws IOException 
    {
       output_ = output;
       BitUtils.WriteString(output_, "GIF89a");
        
       ScreenDescriptor sd = new ScreenDescriptor(width_, height_,
                                                   numColors_);
       sd.Write(output_);

       output_.write(colors_, 0, colors_.length); //Write the Global Color Table
        
       if (JQArgs != null) {
         // Write JavaQuat Application_Extension.
         byte [] authCode = new byte[3];
         authCode[0] = (byte)'0';
         authCode[1] = (byte)'2';
         authCode[2] = (byte)'0';
         ApplicationExtension JQExt = 
           new ApplicationExtension("JavaQuat", authCode);
         JQExt.write(output, JQArgs);
       }
        
//        // Control Transparency with Graphics Control Extension.
//        GraphicsControlExtension gce = new GraphicsControlExtension(-1); //-1 => no transp.
//        gce.Write(output);

       // TABLE BASED IMAGE
       if (transparentPixelIndex >= 0) {
         GraphicsControlExtension gce = new GraphicsControlExtension(transparentPixelIndex);
         gce.write(output_);
       }
       ImageDescriptor id = new ImageDescriptor(width_, height_, ',');
       if (interlaced_p) {
         id.SetInterlaceFlag((byte)1);
       }
       id.Write(output_);

       byte codesize = BitUtils.BitsNeeded(numColors_);
       if (codesize == 1)
           ++codesize;
       output_.write(codesize); // LZW minimum code size
       
       LZWCompressor.writeHeader(output_, codesize);
     }
     
     public void javaQuatWritePixels (byte[] pixels, int first, int bound_exclusive)
       throws IOException  /* NEW */
     {
       LZWCompressor.writePixels(pixels, first, bound_exclusive);
     }


     public void javaQuatWriteFooter () throws IOException  /* NEW */
     {
        LZWCompressor.writeFooter();
        output_.write(0);        // LZW Block Terminator

//        id = new ImageDescriptor((byte)0, (byte)0, ';');
//        id.Write(output);
        output_.write((byte)0x3B);  // GIF Trailer
        output_.flush();
    }
    
}

class BitFile 
{
    OutputStream output_;
    byte buffer_[];
    int index_, bitsLeft_;

    public BitFile(OutputStream output) {
        output_ = output;
        buffer_ = new byte[256];
        index_ = 0;
        bitsLeft_ = 8;
    }

    public void Flush() throws IOException {
        int numBytes = index_ + (bitsLeft_ == 8 ? 0 : 1);
        if (numBytes > 0) {
            output_.write(numBytes);
            output_.write(buffer_, 0, numBytes);
            buffer_[0] = 0;
            index_ = 0;
            bitsLeft_ = 8;
        }
    }

    public void WriteBits(int bits, int numbits) throws IOException {
        int bitsWritten = 0;
        int numBytes = 255;
        do {
            if ((index_ == 254 && bitsLeft_ == 0) || index_ > 254) {
                output_.write(numBytes);
                for(int i = 0; i < numBytes; i++) {
                  output_.write(buffer_[i]);
                }
                output_.flush();
                // SUSPECT CODE, since the similar read doesn't work.
                // output_.write(buffer_, 0, numBytes);

                buffer_[0] = 0;
                index_ = 0;
                bitsLeft_ = 8;
            }

            if (numbits <= bitsLeft_) {
                buffer_[index_] |= (bits & ((1 << numbits) - 1)) <<
                    (8 - bitsLeft_);
                bitsWritten += numbits;
                bitsLeft_ -= numbits;
                numbits = 0;
            }
            else {
                buffer_[index_] |= (bits & ((1 << bitsLeft_) - 1)) <<
                    (8 - bitsLeft_);
                bitsWritten += bitsLeft_;
                bits >>= bitsLeft_;
                numbits -= bitsLeft_;
                buffer_[++index_] = 0;
                bitsLeft_ = 8;
            }
        } while (numbits != 0);
    }
}

class LZWStringTable {
    private final static int RES_CODES = 2;
    private final static short HASH_FREE = (short)0xFFFF;
    private final static short NEXT_FIRST = (short)0xFFFF;
    private final static int MAXBITS = 12;
    private final static int MAXSTR = (1 << MAXBITS);
    private final static short HASHSIZE = 9973;
    private final static short HASHSTEP = 2039;

    byte strChr_[];
    short strNxt_[];
    short strHsh_[];
    short numStrings_;

    public LZWStringTable() {
        strChr_ = new byte[MAXSTR];
        strNxt_ = new short[MAXSTR];
        strHsh_ = new short[HASHSIZE];    
    }

    public int AddCharString(short index, byte b) {
        int hshidx;

        if (numStrings_ >= MAXSTR)
            return 0xFFFF;
        
        hshidx = Hash(index, b);
        while (strHsh_[hshidx] != HASH_FREE)
            hshidx = (hshidx + HASHSTEP) % HASHSIZE;
        
        strHsh_[hshidx] = numStrings_;
        strChr_[numStrings_] = b;
        strNxt_[numStrings_] = (index != HASH_FREE) ? index : NEXT_FIRST;

        return numStrings_++;
    }
    
    public short FindCharString(short index, byte b) {
        int hshidx, nxtidx;

        if (index == HASH_FREE)
            return (short)(b < 0 ? b + 256 : b);  // FIX ***** <<<<<<

        hshidx = Hash(index, b);
        while ((nxtidx = strHsh_[hshidx]) != HASH_FREE) {
            if (strNxt_[nxtidx] == index && strChr_[nxtidx] == b)
                return (short)nxtidx;
            hshidx = (hshidx + HASHSTEP) % HASHSIZE;
        }

        return (short)0xFFFF;
    }

    public void ClearTable(int codesize) {
        numStrings_ = 0;
        
        for (int q = 0; q < HASHSIZE; q++) {
            strHsh_[q] = HASH_FREE;
        }

        int w = (1 << codesize) + RES_CODES;
        for (int q = 0; q < w; q++)
            AddCharString((short)0xFFFF, (byte)q);
    }
    
    static public int Hash(short index, byte lastbyte) {
        return ((int)((short)(lastbyte << 8) ^ index) & 0xFFFF) % HASHSIZE;
    }
}

class LZWCompressor 
{
  private static OutputStream output_;  /* NEW */
  private static int codesize_;   /* NEW */
  private static byte c_;
  private static short index_;
  private static int clearcode_, endofinfo_, numbits_, limit_, errcode_;
  private static short prefix_ = (short)0xFFFF;
  private static BitFile bitFile_;
  private static LZWStringTable strings_;

  public static void LZWCompress(OutputStream output, int codesize,
                                   byte toCompress[]) throws IOException 
  {
//        byte c;
//        short index;
//        int clearcode, endofinfo, numbits, limit, errcode;
//        short prefix = (short)0xFFFF;
     output_ = output;
     codesize_ = codesize;

     bitFile_ = new BitFile(output);
     strings_ = new LZWStringTable();

     clearcode_ = 1 << codesize;
     endofinfo_ = clearcode_ + 1;
    
     numbits_ = codesize + 1;
     limit_ = (1 << numbits_) - 1;
        
     strings_.ClearTable(codesize);
     bitFile_.WriteBits(clearcode_, numbits_);

     for (int loop = 0; loop < toCompress.length; ++loop) 
     {
       c_ = toCompress[loop];
//       System.out.println("LZWCompress pixel = " + c_);
       if ((index_ = strings_.FindCharString(prefix_, c_)) != -1) {
         prefix_ = index_;
       } else {
         bitFile_.WriteBits(prefix_, numbits_);
//         System.out.println("LZWCompress written code = " + prefix_ + " bits= " + numbits_);
         if (strings_.AddCharString(prefix_, c_) > limit_) {
           if (++numbits_ > 12) {
             bitFile_.WriteBits(clearcode_, numbits_ - 1);
             strings_.ClearTable(codesize_);
             numbits_ = codesize_ + 1;
           }
           limit_ = (1 << numbits_) - 1;
         }
         prefix_ = (short)(c_ < 0 ? c_ + 256 : c_);  // FIX **** <<<<<<<<
       }
     }
        
     if (prefix_ != -1)
       bitFile_.WriteBits(prefix_, numbits_);
        
     bitFile_.WriteBits(endofinfo_, numbits_);
     bitFile_.Flush();
   }
    
   public static void writeHeader (OutputStream output, int codesize) throws IOException  /* NEW */
   {
     output_ = output;
     codesize_ = codesize;

     bitFile_ = new BitFile(output);
     strings_ = new LZWStringTable();

     clearcode_ = 1 << codesize;
     endofinfo_ = clearcode_ + 1;
    
     numbits_ = codesize + 1;
     limit_ = (1 << numbits_) - 1;
     prefix_ = (short)0xFFFF;
        
     strings_.ClearTable(codesize);
     bitFile_.WriteBits(clearcode_, numbits_);
   }
    
   public static void writePixels (byte[] pixels, int first, int bound_exclusive)
     throws IOException  /* NEW */
   {
     for (int loop = first; loop < bound_exclusive; ++loop) 
     {
       c_ = pixels[loop];
       if ((index_ = strings_.FindCharString(prefix_, c_)) != -1) {
         prefix_ = index_;
       } else {
         bitFile_.WriteBits(prefix_, numbits_);
         if (strings_.AddCharString(prefix_, c_) > limit_) {
           if (++numbits_ > 12) {
             bitFile_.WriteBits(clearcode_, numbits_ - 1);
             strings_.ClearTable(codesize_);
             numbits_ = codesize_ + 1;
           }
           limit_ = (1 << numbits_) - 1;
         }
         //if (debug_p) { System.out.print(">"); } // $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$  
         prefix_ = (short)(c_ < 0 ? c_ + 256 : c_);  // FIX **** <<<<<<<<
       }
     }
   }
    
   public static void writeFooter () throws IOException  /* NEW */
   {
     if (prefix_ != -1)
       bitFile_.WriteBits(prefix_, numbits_);
        
     bitFile_.WriteBits(endofinfo_, numbits_);
     bitFile_.Flush();
     
     // We are finished and need to clean up for System.gc()
     bitFile_ = null;
     strings_ = null;
     System.gc();
   }
}

class ScreenDescriptor {
    public short localScreenWidth_, localScreenHeight_;
    private byte byte_;  // these are the packed bits, flags.
    public byte backgroundColorIndex_, pixelAspectRatio_;

    public ScreenDescriptor(short width, short height, int numColors) {
        localScreenWidth_ = width;
        localScreenHeight_ = height;
        SetGlobalColorTableSize((byte)(BitUtils.BitsNeeded(numColors) - 1));
        SetGlobalColorTableFlag((byte)1);  // 1 => global color table is present.
        SetSortFlag((byte)0);              // 0 => color table is not sorted
        SetColorResolution((byte)7);       // 7 => 8 bit color
        backgroundColorIndex_ = 0;         // color index 0 is background
        pixelAspectRatio_ = 0;             // meaning no ratio given
    }

    public void Write(OutputStream output) throws IOException {
        BitUtils.WriteWord(output, localScreenWidth_);
        BitUtils.WriteWord(output, localScreenHeight_);
        output.write(byte_);
        output.write(backgroundColorIndex_);
        output.write(pixelAspectRatio_);
    }

    public void SetGlobalColorTableSize(byte num) {
        byte_ |= (num & 7);
    }

    public void SetSortFlag(byte num) {
        byte_ |= (num & 1) << 3;
    }

    public void SetColorResolution(byte num) {
        byte_ |= (num & 7) << 4;
    }
    
    public void SetGlobalColorTableFlag(byte num) {
        byte_ |= (num & 1) << 7;
    }
}


class GraphicsControlExtension
{
  private byte extension_indicator = (byte)0x21;
  private byte graphics_control_label = (byte)0xF9;
  private byte block_size = (byte)4;
  private byte flags = (byte)0;
  private short time_delay = 0;
  private byte transparent_index = (byte)0;
  private byte block_terminator = (byte)0;
  
  GraphicsControlExtension (int transp_index)
  {
    if (transp_index >= 0) { // We want a transparent index.
      flags |= 1;
      transparent_index = (byte)transp_index;
    }
  }
  
  public void write (OutputStream output) throws IOException
  {
    output.write(extension_indicator);
    output.write(graphics_control_label);
    output.write(block_size);
    output.write(flags);
    BitUtils.WriteWord(output, time_delay);
    output.write(transparent_index);
    output.write(block_terminator);
  }
}


class ImageDescriptor {
    public byte separator_;
    public short leftPosition_, topPosition_, width_, height_;
    private byte byte_;

    public ImageDescriptor(short width, short height, char separator) {
        separator_ = (byte)separator;
        leftPosition_ = 0;
        topPosition_ = 0;
        width_ = width;
        height_ = height;
        SetLocalColorTableSize((byte)0);
        SetReserved((byte)0);
        SetSortFlag((byte)0);
        SetInterlaceFlag((byte)0);
        SetLocalColorTableFlag((byte)0);
    }
    
    public void Write(OutputStream output) throws IOException {
        output.write(separator_);
        BitUtils.WriteWord(output, leftPosition_);
        BitUtils.WriteWord(output, topPosition_);
        BitUtils.WriteWord(output, width_);
        BitUtils.WriteWord(output, height_);            
        output.write(byte_);
    }

    public void SetLocalColorTableSize(byte num) {
        byte_ |= (num & 7);
    }

    public void SetReserved(byte num) {
        byte_ |= (num & 3) << 3;
    }

    public void SetSortFlag(byte num) {
        byte_ |= (num & 1) << 5;
    }
    
    public void SetInterlaceFlag(byte num) {
        byte_ |= (num & 1) << 6;
    }

    public void SetLocalColorTableFlag(byte num) {
        byte_ |= (num & 1) << 7;
    }
}


//            new ApplicationExtension("JAVAQUAT", authCode);
//          JQExt.write(output, JQArgs);
class ApplicationExtension 
{
  private String identifier_;
  private byte [] authCode_;
  
  ApplicationExtension (String identifier, byte [] authCode)
  {
    identifier_ = new String(identifier);
    authCode_ = new byte[3];
    System.arraycopy(authCode, 0, authCode_, 0, 3);
  }
  
  void write (OutputStream output, String data) throws IOException 
  {
    output.write((byte)0x21);  // Application Extension Introducer
    output.write((byte)0xFF);  // Application Extension Label
    output.write((byte)11);    // Block Size
    BitUtils.WriteString(output, identifier_);
    output.write(authCode_[0]);
    output.write(authCode_[1]);
    output.write(authCode_[2]);
    
    // Write Data
    int writtenIndex = 0;
    int remaining = data.length();
    while (remaining > 0) {
      int blockSize = Math.min(remaining, 255);
      output.write((byte)blockSize); // Length of block;
      int bound = writtenIndex + blockSize;
      while (writtenIndex < bound) {
        output.write((byte)(data.charAt(writtenIndex++)));
      }
      remaining -= blockSize;
    }
    
    // Write Block Terminator
    output.write((byte)0);
  }
}

class BitUtils {
    public static byte BitsNeeded(int n) {
        byte ret = 1;

        if (n-- == 0)
            return 0;

        while ((n >>= 1) != 0)
            ++ret;
        
        return ret;
    }    

    public static void WriteWord(OutputStream output,
                                 short w) throws IOException {
        output.write(w & 0xFF);
        output.write((w >> 8) & 0xFF);
    }
    
    static void WriteString(OutputStream output,
                            String string) throws IOException {
        for (int loop = 0; loop < string.length(); ++loop)
            output.write((byte)(string.charAt(loop)));
    }
}
