/*
 * Thanks to Adam Doppelt who showed me, through his GIF code, that I could do this.
 */
package JQJoiner.TIFFEncoderPkg;

import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.util.*;

/*
 * 1. Use this by creating a TIFFEncoder.  There will be one for palette/256 color images and
 * one for RGB millions of colors images (Direct Color).
 * 2. If you want resolution other than 72 ppi call setResolution.
 * 3. Call write, passing stream and some pixels.  Repeat as necessary.
 *    Write will close the file and return true when  (width * height) pixels have been written.
 */


public class TIFFEncoder
{
  // Compression type.
  public final static short NO_COMPRESSION = 1;
  public final static int LZW = 5; // Not implemented. Readers are not required to understand it.
  public final static int PACKBITS = 32773;
  
  public final static int DO_IN_MEMORY_MAX = 5000000; //Pictures this size or smaller done in memory.

  // Parameters
  boolean tiledP_ = false;
  int compressionType_ = PACKBITS; //NO_COMPRESSION;
  int numberOfFields_; // in the IFD.
      

  // For color model.
  private final static int PALETTE = 3;
  private final static int DIRECT_COLOR = 2;
  // For resolution.
  private final static int INCH = 2;
  private final static int CENTIMETER = 3;
  // Used in the header.
  private final static short LITTLEENDIAN = 0x4949;
  private final static short BIGENDIAN = 0x4D4D;
  private final static short TIFFCODE = 42;
  // Used to compute offset_.
  private final static short WORD_LENGTH = 4;
  private final static short SHORT_LENGTH = 2;
  private final static short LONG_LENGTH = 4;
  private final static short RATIONAL_LENGTH = 8;
  private final static short HEADER_LENGTH = 8;
  private final static short FIELD_LENGTH = 12;
  private final static short NumFeildsSIZE = 2;
  private final static short NextIFDOffsetSize = 4;

  // Field (Directory Entry) Tags.
  private final static short IMAGE_WIDTH = 256;
  private final static short IMAGE_LENGTH = 257;
  private final static short BITS_PER_SAMPLE = 258;
  private final static short COMPRESSION = 259;
  private final static short PHOTOMETRIC_INTERPRETATION = 262;
  private final static short STRIP_OFFSETS = 273;
  private final static short ROWS_PER_STRIP = 278;
  private final static short STRIP_BYTE_COUNTS = 279;
  private final static short X_RESOLUTION = 282;
  private final static short Y_RESOLUTION = 283;
  private final static short RESOLUTION_UNIT = 296;
  private final static short COLOR_MAP = 320;
  private final static short TILE_WIDTH = 322;
  private final static short TILE_LENGTH = 323;
  private final static short TILE_OFFSETS = 324;
  private final static short TILE_BYTE_COUNTS = 325;
  private final static int   COPYRIGHT = 33432;
  
  // Field Types
  private final static short BYTE = 1;
  private final static short ASCII = 2;
  private final static short SHORT = 3;
  private final static short LONG = 4;
  private final static short RATIONAL = 5;

  // Class variables
  private boolean littleEndian_ = false;
  private int offset_ = 0;
  private int width_;
  private int height_;
  private IndexColorModel colorMap_;
  private int ppi_;
  private int numColors_;
  private int bitsPerPixel_;
  private int pixelsProcessed_;
  private int photometricInterpretation_;
  private byte [] reds_;
  private byte [] greens_;
  private byte [] blues_;

  // For the compressed pixel data.
  private boolean pixelsInOneGroupP_;
  private int firstPixelsOffset_;
  private int stripIndex_;
  private int [] byteCounts_;
  private int tileWidth_;
  private int tileLength_;
  
  private int tileRows_;
  private int tileColumns_;

//  private int pixelOffsets_offset_;
//  private int byteCounts_offset_;

  private String tempCompressionPathname_ = null;
  private OutputStream tempCompressionStream_ = null;

  private byte [] pixels_ = null;
  private int rowsPerStrip_ = 0;
  private int strips_;
  private byte [] [] pixelAccumulator_ = null;
  private int accRowIndex_ = 0;
  private int accColIndex_ = 0;
  private boolean storeInMemory_;
  
  private String copyrightNotice_ = null;
  private boolean writeHasBegunP_ = false;
  private boolean verbose_ = false;

  public TIFFEncoder (int width, int height, IndexColorModel colorModel)
  {
    width_ = width;
    height_ = height;
    stripIndex_ = 0;
    if (width_ * height_ <= DO_IN_MEMORY_MAX) {
      storeInMemory_ = true;
    } else {
      storeInMemory_ = false;
    }
    System.out.println("TIFFEncoder.constructor: storeInMemory_=" + storeInMemory_);
    if (tiledP_) {
      setTileParameters();
      byteCounts_ = new int [tileRows_ * tileColumns_];
      pixelsInOneGroupP_ = tileRows_ * tileColumns_ == 1;
      numberOfFields_ = 13;
    } else {
      //setStripParameters(height_);
      numberOfFields_ = 12;
    }
    numColors_ = colorModel.getMapSize();
    if (numColors_ <= 16) {
      bitsPerPixel_ = 4;
    } else {
      bitsPerPixel_ = 8;
    }
    reds_ = new byte[numColors_];
    greens_ = new byte[numColors_];
    blues_ = new byte[numColors_];
    colorModel.getReds(reds_);
    colorModel.getGreens(greens_);
    colorModel.getBlues(blues_);
    ppi_ = 72;
    pixelsProcessed_ = 0;
    photometricInterpretation_ = PALETTE;
    tempCompressionPathname_ = "JQTiffBytes";
  }
  
  public void setVerbose (boolean value)
  {
    //System.out.println("TIFFEncoder.setVerbose Entry: value=" + value);
    verbose_ = value;
  }
  
  public void setResolution (int ppi)
  {
    //System.out.println("TIFFEncoder.setResolution Entry: ppi=" + ppi);
    ppi_ = ppi;
  }
  
  public void setStripParameters (int rowsPerStrip)
    // Arg of 0 means one strip.
  {
    if (writeHasBegunP_) {
      System.err.println("Too late to set rowsPerStrip, caching has already started.");
      return;
    }
    if (rowsPerStrip <= 0) {
      rowsPerStrip = height_;
    }
    rowsPerStrip_ = rowsPerStrip;
    strips_ = (int)Math.ceil(height_ / (float)rowsPerStrip_); // The general expression for this.
    pixelAccumulator_ = new byte [rowsPerStrip_] [width_];
    byteCounts_ = new int[strips_];
    pixelsInOneGroupP_ = strips_ == 1;
    accRowIndex_ = 0;
    accColIndex_ = 0;
  }
  
  public void setCompression (int compression)
  {
    if (writeHasBegunP_) {
      System.err.println("Too late to set compression type, caching has already started.");
      return;
    }
    if (compression == NO_COMPRESSION || compression == PACKBITS) {
      compressionType_ = compression;
    } else {
      System.err.println("Tiff Encoder: Attempt to set compression type to " + compression);
    }
  }
  
  public void setCopyrightNotice (String notice)
  { 
    if (notice == null || notice.length() == 0) {
      copyrightNotice_ = null;
      if (tiledP_) {
        numberOfFields_ = 13;
      } else {
        numberOfFields_ = 12;
      }
    } else {
      copyrightNotice_ = notice;
      if (tiledP_) {
        numberOfFields_ = 14;
      } else {
        numberOfFields_ = 13;
      }
    }
  }

  public boolean write (OutputStream output, byte[] pixels, int startIndex, int bytes) 
         throws AWTException, IOException
    // If we are doing compression we do it to a temp file because TIFF format requires
    // that we write offsets to data before we know what they are.  This is not the best
    // design for a file format one can imagine.
    // In this encoder we will write the HEADER, OFFSET data, IFD, and PIXEL data.
    // Note that OFFSET data and the IFD need to be on word (4 byte) boundaries.
  {
    if (rowsPerStrip_ <= 0) {
      if (width_ * height_ > DO_IN_MEMORY_MAX) {
        int rps = DO_IN_MEMORY_MAX / width_;
        if (rps == 0)
          rps = 1;
        setStripParameters(rps);
      } else {
        setStripParameters(height_);
      }
    }
    writeHasBegunP_ = true;
    // Parameters for writing
    if (pixels != null && pixels.length >= startIndex + bytes) {
      addPixels(pixels, compressionType_, startIndex, bytes);
      pixelsProcessed_ += bytes;
    } else if (pixels != null) {
      System.err.println("TIFFEncoder.write received pixel array too short for start + length");
    }
    
    if (pixelsProcessed_ >= width_ * height_) {
      // Store any fractional strip.
      if (accColIndex_ > 0 || accRowIndex_ > 0) {
        storeStrip(compressionType_);
      }
      //++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
      if (!storeInMemory_ && tempCompressionStream_ != null) {
        // Close the temp file.
        tempCompressionStream_.flush();
        tempCompressionStream_.close();
      }
      // Write offset data and figure out offsets.
      int xres_offset = HEADER_LENGTH;
      int yres_offset = xres_offset + RATIONAL_LENGTH;
      int pixelOffsets_offset = yres_offset + RATIONAL_LENGTH;
      int byteCounts_length = 0;
      if (! pixelsInOneGroupP_) {
        if (tiledP_) {
          byteCounts_length = LONG_LENGTH * tileRows_ * tileColumns_;
        } else {
          byteCounts_length = LONG_LENGTH * strips_;
        }
      }
      int byteCounts_offset = pixelOffsets_offset + byteCounts_length;
      int copyrightOffset = byteCounts_offset + byteCounts_length;
      int color_map_offset = copyrightOffset + 
        (copyrightNotice_ == null ? 0 : (int)Math.max(5, copyrightNotice_.length() + 1));
      int IFD_offset = color_map_offset + colorMapBytes(bitsPerPixel_);
      boolean padBeforeIFD = (IFD_offset % WORD_LENGTH > 0);
      if (padBeforeIFD) {
        IFD_offset += 2; // No single bytes have been written to this point, only 2 byte shorts.
      }
      int firstPixelsOffset_ = IFD_offset + ifdLength(numberOfFields_);
      
      // Write header and offset data
      writeHeader(output, IFD_offset);
      writeRational(output, ppi_, 1);
      writeRational(output, ppi_, 1);
      if (! pixelsInOneGroupP_) {
        writePixelsOffsetData(output, firstPixelsOffset_, byteCounts_);
        writeByteCountData(output, byteCounts_);
      }
      if (copyrightNotice_ != null) {
        writeCopyrightData (output, copyrightNotice_);
      }
      writeColors(output, numColors_, bitsPerPixel_, reds_, greens_, blues_);
      if (padBeforeIFD) {
        writeShort(output, 0);
      }

      // Write IFD
      writeShort(output, numberOfFields_);
      writeImageWidth(output, width_);                               /* 1 */
      writeImageHeight(output, height_);                             /* 2 */
      writeBitsPerPixel(output, bitsPerPixel_);                      /* 3 */
      writeCompression(output, compressionType_);                    /* 4 */
      writePhotometric(output, photometricInterpretation_);          /* 5 */
      int pixOffset = (pixelsInOneGroupP_ ? firstPixelsOffset_ : pixelOffsets_offset);
      int pixBytes = (pixelsInOneGroupP_ ? 
                      byteCounts_[0] :
                      byteCounts_offset);
      if (!tiledP_) {
        writeStripOffsets(output, byteCounts_.length, pixOffset);    /* 10 */
        writeRowsPerStrip(output, rowsPerStrip_);                    /* 11 */
        writeStripByteCounts(output, byteCounts_.length, pixBytes);  /* 12 */
      }
      writeXResolution(output, xres_offset);                         /* 6 */
      wrireYResolution(output, yres_offset);                         /* 7 */
      writeResolutionUnit(output, INCH);                             /* 8 */
      writeColorMap(output, colorMapLength(bitsPerPixel_), color_map_offset);   /* 9 */
      if (tiledP_) {
        writeTileWidth (output, tileWidth_);                         /* 10 */
        writeTileHeight (output, tileLength_);                       /* 11 */
        writeTileOffsets (output, byteCounts_.length, pixOffset);    /* 12 */
        writeTileByteCounts (output, byteCounts_.length, pixBytes);  /* 13 */
      }
      if (copyrightNotice_ != null) {
        writeCopyrightField (output, copyrightNotice_, copyrightOffset);
      }
      writeLong(output,0); //No Next IFD;
      
      // Write pixels
      writeCompressedPixels(output, tempCompressionPathname_);
      writeHasBegunP_ = false;
      return true;
    }
    return false;
  }
  
  /* *************************** SIZE UTILITIES *************************** */
  
  private void setTileParameters ()
  {
    if (writeHasBegunP_) {
      System.err.println("Too late to set tile parameters, caching has already started.");
      return;
    }
    tileWidth_ = width_;
    tileLength_ = height_;
    tileRows_ = 1;
    tileColumns_ = 1;
  }
  
  private int colorMapLength (int bitsPerPixel)
  {
    return 3 * (int)Math.pow(2, bitsPerPixel);  // 3 = colors in RGB.
  }
  
  private int colorMapBytes (int bitsPerPixel)
  {
    return SHORT_LENGTH * colorMapLength(bitsPerPixel);
  }
  
  private int ifdLength (int fields)
  {
    return NumFeildsSIZE + fields * FIELD_LENGTH + NextIFDOffsetSize;
  }
  
  /* ********************** PIXEL accumulation, compression, and writing ******************* */

  private void addPixels(byte [] pixels, int compressionType, int startIndex, int bytes) 
         throws IOException
     // This forms the strips and calls storeStrip to compress pixels, save them, and set
     // byteCounts_ elements.  When we finish the picture the caller will call storeStrip
     // or the equivalent if needed.
  {
    for (int pixIndex = startIndex; pixIndex < startIndex + bytes; pixIndex++) {
      pixelAccumulator_[accRowIndex_][accColIndex_++] = pixels[pixIndex];
      if (accColIndex_ >= width_) {
        accColIndex_ = 0;
        accRowIndex_++;
        if (accRowIndex_ >= rowsPerStrip_) {
          accRowIndex_ = 0;
          storeStrip(compressionType);
        }
      }
    }
  }

  private void storeStrip (int compressionType) 
         throws IOException
  {
    //System.out.println("TIFFEncoder.storeStrip pt 1");
    int bytesPerStrip = width_ * rowsPerStrip_;
    int maxRow = rowsPerStrip_;
    if (height_ - stripIndex_ * rowsPerStrip_ < rowsPerStrip_) {
      maxRow = height_ - stripIndex_ * rowsPerStrip_;
    }
    if (compressionType == NO_COMPRESSION) {
      byteCounts_[stripIndex_] = bytesPerStrip;
    } else {
      byteCounts_[stripIndex_] = 0; // It will be incremented as we compress the pixel data.
    }
    if (storeInMemory_) {
      // Store the strip into the array pixels.
      if (pixels_ == null) {
        pixels_ = new byte [strips_ * bytesPerStrip];
      }
      int pixIndex = 0; // Serves as an index into the array of possibly compressed pixels_.
      if (compressionType == NO_COMPRESSION) {
        pixIndex = stripIndex_ * bytesPerStrip;
      } else if (compressionType == PACKBITS) {
        for (int i = 0; i < stripIndex_; i++) {
          pixIndex += byteCounts_[i];
        }
      }
      for (int row = 0; row < maxRow; row++) {
        if (compressionType == NO_COMPRESSION) {
          for (int col = 0; col < width_; col++) {
            pixels_[pixIndex++] = pixelAccumulator_[row][col];
          }
          // byteCounts_[stripIndex_] is set above.
        } else if (compressionType == PACKBITS) {
          int packedBytes = packBitsIntoMemory(pixelAccumulator_[row], pixIndex);
          byteCounts_[stripIndex_] += packedBytes;
          pixIndex += packedBytes;
        }
      }
    } else {  // If not in memory then save to temp file.
      if (tempCompressionStream_ == null) {
        getTempOutputStream();
      }
      if (tempCompressionStream_ == null) {
        throw new IOException("Tiff Encoder failed to create a work file.");
      }
      for (int row = 0; row < maxRow; row++) {
        if (compressionType == NO_COMPRESSION) {
          for (int col = 0; col < width_; col++) {
            tempCompressionStream_.write(pixelAccumulator_[row][col]);
          }
          if (verbose_ && row % 100 == 0) {
            System.out.println("Cached row " + row + " strip " + stripIndex_);
          }
        } else if (compressionType == PACKBITS) {
          int packedBytes = packBitsIntoFile(pixelAccumulator_[row]);
          if (verbose_ && row % 100 == 0) {
            System.out.println("Packed Bits of row " + row + " strip " + stripIndex_);
          }
          byteCounts_[stripIndex_] += packedBytes;
        }
      }
    }
    stripIndex_++;
  }
  
  private int packBitsIntoFile (byte [] row) throws IOException 
    // Returns the number of bytes after compression.  THIS is PACKBITS compression.
  {
    int runStart = 0;
    int previous = 0;
    int current = 1;
    int bytesOut = 0;
    while (current < width_) {
      // Handle repeat runs.
      byte runStartByte = row[runStart];
      while (current < width_ && runStartByte == row[current]) {
        current++;
      } // current is no bigger than width_
      if (current - runStart > 1) {
        while (current - runStart >= 128) {
          tempCompressionStream_.write(-((byte)127)); // count is 1 less than repetitions.
          tempCompressionStream_.write(runStartByte);
          runStart += 128;
          bytesOut += 2;
        }
        // Current is the index of the first non-matching byte or it is width_;
        if (current - runStart > 1) {
          tempCompressionStream_.write((byte)(runStart - current + 1));// count is 1 less than repetitions.
          tempCompressionStream_.write(runStartByte);
          runStart = current;
          bytesOut += 2;
        }
        current = runStart + 1;
      }
      // Handle literal (i.e.no repeat) case.
      previous = runStart;
      while (current < width_ && (row[previous] != row[current] ||
                                  current + 1 < width_ && row[current + 1] != row[current]) ) {
        previous = current;
        current++;
      } // current is no bigger than width_
      int bound = (current >= width_ ? width_ : previous); //bound is first byte not used here.
      if (bound - runStart > 0) {
        while (current - runStart > 128) {
          tempCompressionStream_.write((byte)127); // count is 1 less than string of bytes length.
          for (int i = 0; i < 128; i++) {
            tempCompressionStream_.write(row[runStart++]);
          }
          bytesOut += 129;
        }
        // 2 cases exist: bound - runStart > 0, == 0
        int remainingBytes = bound - runStart;
        if (remainingBytes > 0) {
          tempCompressionStream_.write((byte)(remainingBytes - 1)); // count is 1 less than string length.
          for (int i = 0; i < remainingBytes; i++) {
            tempCompressionStream_.write(row[runStart++]);
          }
          bytesOut += remainingBytes + 1;
        }
      }
    }
    if (runStart == width_ - 1) {
      tempCompressionStream_.write((byte)1);
      tempCompressionStream_.write(row[runStart]);
      bytesOut += 2;
    }
    return bytesOut;
  }
 
  private int packBitsIntoMemory (byte [] row, int pixIndex)
    // Returns the number of bytes after compression.  THIS is PACKBITS compression.
  {
    int runStart = 0;
    int previous = 0;
    int current = 1;
    int bytesOut = 0;
    while (current < width_) {
      // Handle repeat runs.
      byte runStartByte = row[runStart];
      while (current < width_ && runStartByte == row[current]) {
        current++;
      } // current is no bigger than width_
      if (current - runStart > 1) {
        while (current - runStart >= 128) {
          pixels_[pixIndex++] = -((byte)127); // count is 1 less than repetitions.
          pixels_[pixIndex++] = runStartByte;
          runStart += 128;
          bytesOut += 2;
        }
        // Current is the index of the first non-matching byte or it is width_;
        if (current - runStart > 1) {
          pixels_[pixIndex++] = (byte)(runStart - current + 1);// count is 1 less than repetitions.
          pixels_[pixIndex++] = runStartByte;
          runStart = current;
          bytesOut += 2;
        }
        current = runStart + 1;
      }
      // Handle literal (i.e.no repeat) case.
      previous = runStart;
      while (current < width_ && (row[previous] != row[current] ||
                                  current + 1 < width_ && row[current + 1] != row[current]) ) {
        previous = current;
        current++;
      } // current is no bigger than width_
      int bound = (current >= width_ ? width_ : previous); //bound is first byte not used here.
      if (bound - runStart > 0) {
        while (current - runStart > 128) {
          pixels_[pixIndex++] = (byte)127; // count is 1 less than string of bytes length.
          for (int i = 0; i < 128; i++) {
            pixels_[pixIndex++] = row[runStart++];
          }
          bytesOut += 129;
        }
        // 2 cases exist: bound - runStart > 0, == 0
        int remainingBytes = bound - runStart;
        if (remainingBytes > 0) {
          pixels_[pixIndex++] = (byte)(remainingBytes - 1); // count is 1 less than string length.
          for (int i = 0; i < remainingBytes; i++) {
            pixels_[pixIndex++] = row[runStart++];
          }
          bytesOut += remainingBytes + 1;
        }
      }
    } 
    if (runStart == width_ - 1) {
      pixels_[pixIndex++] = (byte)1;
      pixels_[pixIndex++] = row[runStart];
      bytesOut += 2;
    }
    return bytesOut;
  }
 
  private void writeCompressedPixels(OutputStream output, String tempCompressionPathname)
          throws IOException 
  {
    if (storeInMemory_) {
      System.out.println("Writing pixels in memory to file.");
      int i = 0;
      for (int strip = 0; strip < strips_; strip++) {
        for (int stripByte = 0; stripByte < byteCounts_[strip]; stripByte++) {
          output.write(pixels_[i++]);
        }
      }
    } else {
      System.out.println("Writing pixels in temp file to tiff file.");
      InputStream stream1 = null;
      try {
        stream1 = getTempInputStream();
        int i = 0;
        for (int strip = 0; strip < strips_; strip++) {
          if (verbose_) {
            System.out.println("  Beginning strip " + strip);
          }
          for (int stripByte = 0; stripByte < byteCounts_[strip]; stripByte++) {
            output.write(stream1.read());
          }
        }        
      } catch (IOException e) {
        System.err.println("TIFF Encoder error retrieving pixel data from temp file:" +
                           (e.getMessage() == null ? e.toString() : e.getMessage()) );
      } finally {
        try {output.close();} catch (IOException e){}
        if (stream1 != null) {
          try {stream1.close();} catch (IOException e){}
        }
        try {
          File testExistance = new File(tempCompressionPathname_);
          if (testExistance.exists()) {
            testExistance.delete();
          }
        } catch (Throwable e){}
      }
    }        
    System.out.println("Finished writing pixels.");
  }
  
  private InputStream getTempInputStream () throws IOException
  {
    InputStream stream1 = null;
    try {
      File testExistance = new File(tempCompressionPathname_);
      if (!testExistance.exists()) {
        throw new IOException("File DOES NOT EXIST - no read done.");
      }
      stream1 = new FileInputStream(tempCompressionPathname_);
      stream1 = new BufferedInputStream(stream1);
    } catch (Throwable e) {
      System.err.println("TIFFEncoder.getTempInputStream " + 
                         (e.getMessage() == null ? e.toString() : e.getMessage()) );
      stream1 = null;
    } finally {
      return stream1;
    }
  }

  public OutputStream getTempOutputStream ()
    // This uses the global tempCompressionPathname_ as the basis for creating a
    // OutputStream value for global tempCompressionStream_.  tempCompressionPathname_
    // is updated in the event that the original value didn't work out.
  {
     try {
       File testExistance = new File(tempCompressionPathname_ + ".tmp");
       if (testExistance.exists() && !testExistance.delete()) {
         // Try a new name.  We could check here for filename length.
         // Since Mac only allows 31 characters.
         File tester = null;
         int count = 1;
         String newPathname = tempCompressionPathname_ + count + ".tmp";
         while ((tester = (new File(newPathname))).exists() && !tester.delete()) {
           newPathname = tempCompressionPathname_ + count + ".tmp";
           if (count++ > 1000) {
             throw new IOException("Excessive backup versions exist - no write done.");
           }
         }
         tempCompressionPathname_ = newPathname;
       }
       tempCompressionStream_ = new FileOutputStream(tempCompressionPathname_);
       tempCompressionStream_ = new BufferedOutputStream(tempCompressionStream_);
     } catch (Throwable e) {
       System.err.println("TIFFEncoder.getTempOutputStream " + 
                          (e.getMessage() == null ? e.toString() : e.getMessage()) );
       tempCompressionStream_ = null;
     } finally {
       return tempCompressionStream_;
     }
  }

  /* ********************** Number/String WRITERS ******************* */
  
  private void writeByte(OutputStream output, int w) throws IOException 
  {
    output.write(w & 0xFF);
  }
    
  private void writeShort(OutputStream output, int w) throws IOException 
  {
    if (littleEndian_) {
      output.write(w & 0xFF);
      output.write((w >> 8) & 0xFF);
    } else {
      output.write((w >> 8) & 0xFF);
      output.write(w & 0xFF);
    }
  }
    
  private void writeShort4Bytes(OutputStream output, int w) throws IOException
    // When a short is written in 4 bytes it is always in the first two, regardless
    // of endianness.
  {
    if (littleEndian_) {
      output.write(w & 0xFF);
      output.write((w >> 8) & 0xFF);
      output.write(0);
      output.write(0);
    } else {
      output.write((w >> 8) & 0xFF);
      output.write(w & 0xFF);
      output.write(0);
      output.write(0);
    }
  }
    
  private void writeLong (OutputStream output, int w) throws IOException 
  {
    if (littleEndian_) {
      output.write(w & 0xFF);
      output.write((w >> 8) & 0xFF);
      output.write((w >> 16) & 0xFF);
      output.write((w >> 24) & 0xFF);
    } else {
      output.write((w >> 24) & 0xFF);
      output.write((w >> 16) & 0xFF);
      output.write((w >> 8) & 0xFF);
      output.write(w & 0xFF);
    }
  }
  
  private void writeRational (OutputStream output, int numer, int denom) throws IOException 
  {
    writeLong(output, numer);
    writeLong(output, denom);
  }
    
  private void writeString(OutputStream output, String string) throws IOException 
  {
      for (int loop = 0; loop < string.length(); ++loop)
          output.write((byte)(string.charAt(loop)));
  }

  /* ************** FIELD WRITERS *************** */
               
  private void writeHeader (OutputStream output, int ifdOffset) throws IOException 
  {
    int endian = (littleEndian_ ? LITTLEENDIAN : BIGENDIAN);
    writeShort(output, endian);
    writeShort(output, TIFFCODE);
    writeLong(output, ifdOffset);
  }
  private void writeImageWidth (OutputStream output, int width) throws IOException 
  {
    writeShort(output, IMAGE_WIDTH);
    writeShort(output, LONG);
    writeLong(output, 1);
    writeLong(output, width);
  }
  private void writeImageHeight (OutputStream output, int height) throws IOException 
  {
    writeShort(output, IMAGE_LENGTH);
    writeShort(output, LONG);
    writeLong(output, 1);
    writeLong(output, height);
  }
  private void writeBitsPerPixel (OutputStream output, int bitsPerPixel) throws IOException 
  {
    writeShort(output, BITS_PER_SAMPLE);
    writeShort(output, SHORT);
    writeLong(output, 1);
    writeShort4Bytes(output, bitsPerPixel);
  }
  private void writeCompression (OutputStream output, int compressionType) throws IOException 
  {
    writeShort(output, COMPRESSION);
    writeShort(output, SHORT);
    writeLong(output, 1);
    writeShort4Bytes(output, compressionType);
  }
  private void writePhotometric (OutputStream output, int photometricInterpretation_) throws IOException 
  {
    writeShort(output, PHOTOMETRIC_INTERPRETATION);
    writeShort(output, SHORT);
    writeLong(output, 1);
    writeShort4Bytes(output, photometricInterpretation_);
  }
  
  private void writeStripOffsets (OutputStream output, int count, int value) throws IOException 
  {
    writeShort(output, STRIP_OFFSETS);
    writeShort(output, LONG);  // Meaning the data at the offset is Long.
    writeLong(output, count);
    writeLong(output, value); // Example pg. 20 shows this written as long.
  }
  private void writeRowsPerStrip (OutputStream output, int height) throws IOException 
  {
    writeShort(output, ROWS_PER_STRIP);
    writeShort(output, LONG);
    writeLong(output, 1);
    writeLong(output, height);
  }
  private void writeStripByteCounts (OutputStream output, int count, int value) throws IOException 
  {
    writeShort(output, STRIP_BYTE_COUNTS);
    writeShort(output, LONG);  // Meaning the data at the offset is Long.
    writeLong(output, count);
    writeLong(output, value); // Example pg. 20 shows this written as long.
  }
  
  private void writeXResolution (OutputStream output, int offset) throws IOException 
  {
    writeShort(output, X_RESOLUTION);
    writeShort(output, RATIONAL);  // Meaning the data at the offset is Rational.
    writeLong(output, 1);
    writeLong(output, offset); // Example pg. 20 shows this written as long.
  }
  private void wrireYResolution (OutputStream output, int offset) throws IOException 
  {
    writeShort(output, Y_RESOLUTION);
    writeShort(output, RATIONAL);  // Meaning the data at the offset is Rational.
    writeLong(output, 1);
    writeLong(output, offset); // Example pg. 20 shows this written as long.
  }
  private void writeResolutionUnit (OutputStream output, int unitOfLength) throws IOException 
  {
    writeShort(output, RESOLUTION_UNIT);
    writeShort(output, SHORT);
    writeLong(output, 1);
    writeShort4Bytes(output, unitOfLength);
  }
  private void writeColorMap (OutputStream output, int entries, int offset) throws IOException 
  {
    writeShort(output, COLOR_MAP);
    writeShort(output, SHORT);
    writeLong(output, entries);
    writeLong(output, offset);
  }
  private void writeTileWidth (OutputStream output, int width) throws IOException 
  {
    writeShort(output, TILE_WIDTH);
    writeShort(output, LONG);
    writeLong(output, 1);
    writeLong(output, width);
  }
  private void writeTileHeight (OutputStream output, int length) throws IOException 
  {
    writeShort(output, TILE_LENGTH);
    writeShort(output, LONG);
    writeLong(output, 1);
    writeLong(output, length);
  }
  private void writeTileOffsets (OutputStream output, int count, int offset)
    throws IOException
  {
    writeShort(output, TILE_OFFSETS);
    writeShort(output, LONG);
    writeLong(output, count);
    writeLong(output, offset);
  }
  private void writeTileByteCounts (OutputStream output, int count, int value)
    throws IOException
  {
    writeShort(output, TILE_BYTE_COUNTS);
    writeShort(output, LONG);
    writeLong(output, count);
    writeLong(output, value);
  }
  private void writeCopyrightField (OutputStream output, String notice, int offset)
    throws IOException
  {
    writeShort(output, COPYRIGHT);
    writeShort(output, ASCII);
    byte [] chars = notice.getBytes();
    writeLong(output, (int)Math.max(5, chars.length + 1)); //Make sure it won't fit here.
    writeLong(output, offset);
  }
  
  /* ************************ WRITERS AT OFFSETS ********************* */
  
  private void writeCopyrightData (OutputStream output, String notice) throws IOException 
  {
    byte [] chars = notice.getBytes();
    int i = 0;
    for (; i < chars.length; i++) {
      output.write(chars[i]);
    }
    output.write(0); // Terminator
    i++;
    // Make sure we have at least 5 bytes as we claimed in the IFD field.
    for(; i < 5; i++) {
      output.write(0);
    }
  }

  private void writeColors (OutputStream output, int numColors, int bitsPerPixel, 
                            byte [] reds, byte [] greens, byte [] blues) throws IOException 
     // Our byte values are -128 to 127.  These are converted to 0 to 255.  TIFF values
     // are 0 to 65535 and 65535 = 255 * 257, how convenient.
  {
    int i;
    int bytesRequired = (int)Math.pow(2, bitsPerPixel);
    int Zero_255;
    // Write reds
    for(i = 0; i < numColors; i++) {
      Zero_255 = reds[i];
      if (Zero_255 < 0) Zero_255 += 256;
      writeShort(output, 257 * Zero_255);
    }
    for(; i < bytesRequired; i++) {
      writeShort(output, 0);
    }
    // Write greens
    for(i = 0; i < numColors; i++) {
      Zero_255 = greens[i];
      if (Zero_255 < 0) Zero_255 += 256;
      writeShort(output, 257 * Zero_255);
    }
    for(; i < bytesRequired; i++) {
      writeShort(output, 0);
    }
    // Write blues
    for(i = 0; i < numColors; i++) {
      Zero_255 = blues[i];
      if (Zero_255 < 0) Zero_255 += 256;
      writeShort(output, 257 * Zero_255);
    }
    for(; i < bytesRequired; i++) {
      writeShort(output, 0);
    }
  }

  private void writePixelsOffsetData (OutputStream output, int firstOffset, int [] byteCounts)
          throws IOException 
    // This assumes that the tiles are written consecutively.
  {
    if (byteCounts.length == 1) {
      return;
    }
    int totalOffset = firstOffset;
    try {
      for (int i = 0; i < byteCounts.length; i++) {
        writeLong(output, totalOffset);
        totalOffset += byteCounts[i];
      }
    } catch (ArrayIndexOutOfBoundsException ex) {
      System.err.println("It is impossible for this exception to happen.  " + ex);
    }
  }
  
  private void writeByteCountData (OutputStream output, int [] byteCounts)
          throws IOException 
  {
    if (byteCounts.length == 1) {
      return;
    }
    try {
      for (int i = 0; i < byteCounts.length; i++) {
        writeLong(output, byteCounts[i]);
      }
    } catch (ArrayIndexOutOfBoundsException ex) {
      System.err.println("It is impossible for this exception to happen.  " + ex);
    }
  }
  
}
